@@ -51,6 +51,7 @@ def get_ollama_response(
     messages: List[Dict[str, str]] = None,
     stream: bool = False,
     attachments: List[str] = None,
+    auto_process_tool_calls: bool = False,
     **kwargs,
 ) -> Dict[str, Any]:
     """
@@ -88,11 +89,11 @@ def get_ollama_response(
                         from npcpy.data.load import load_csv
                         csv_data = load_csv(attachment)
                         if csv_data is not None:
-                            csv_sample = csv_data.head(10).to_string()
+                            csv_sample = csv_data.head(100).to_string()
                             if prompt:
-                                prompt += f"\n\nContent from CSV: {os.path.basename(attachment)} (first 10 rows):\n{csv_sample}"
+                                prompt += f"\n\nContent from CSV: {os.path.basename(attachment)} (first 100 rows):\n{csv_sample} \n csv description: {csv_data.describe()}"
                             else:
-                                prompt = f"Content from CSV: {os.path.basename(attachment)} (first 10 rows):\n{csv_sample}"
+                                prompt = f"Content from CSV: {os.path.basename(attachment)} (first 100 rows):\n{csv_sample} \n csv description: {csv_data.describe()}"
                     except Exception:
                         pass
 
@@ -109,6 +110,8 @@ def get_ollama_response(
                 else:
                     messages[-1]["content"].append({"type": "text", "text": prompt})
         else:
+            if not messages:
+                messages = []
             messages.append({"role": "user", "content": prompt})
     if format == "json" and not stream:
         json_instruction = """If you are a returning a json object, begin directly with the opening {.
@@ -135,39 +138,34 @@ def get_ollama_response(
             messages.append({"role": "user", "content": ""})
             last_user_idx = len(messages) - 1
         messages[last_user_idx]["images"] = image_paths
-    
+
     api_params = {
         "model": model,
         "messages": messages,
-        "stream": stream,
+        "stream": stream if not (tools and tool_map and auto_process_tool_calls) else False,
     }
 
-    
     if tools:
         api_params["tools"] = tools
-    if tool_choice:
-        api_params["tool_choice"] = tool_choice
     if think is not None:
         api_params['think'] = think
-    
 
     if isinstance(format, type) and not stream:
         api_params["format"] = format.model_json_schema()
     elif isinstance(format, str) and format == "json" and not stream:
         api_params["format"] = "json"
-    
-
 
     options = {}
     for key, value in kwargs.items():
         if key in [
             "stop", "temperature", "top_p", "max_tokens", "max_completion_tokens",
-            "tools", "tool_choice", "extra_headers", "parallel_tool_calls",
+            "extra_headers", "parallel_tool_calls",
             "response_format", "user",
         ]:
             options[key] = value
+    if tool_choice:
+        options["tool_choice"] = tool_choice
 
-    # Create standardized response structure
     result = {
         "response": None,
         "messages": messages.copy(),
@@ -175,50 +173,125 @@ def get_ollama_response(
         "tool_calls": [], 
         "tool_results": []
     }
-        
-    if stream:
-        result["response"] = ollama.chat(**api_params, options=options)
-        return result
-    
 
-    # Non-streaming case
-    res = ollama.chat(**api_params, options = options)
+    #print("API params:", api_params)
+
+    # If we want raw tool calls OR no tools, just stream directly
+    if not auto_process_tool_calls or not (tools and tool_map):
+        res = ollama.chat(**api_params, options=options)
+        result["raw_response"] = res
+        
+        if stream:
+            result["response"] = res  # This is the stream generator
+            return result
+        else:
+            # Non-streaming regular response
+            message = res.get("message", {})
+            response_content = message.get("content", "")
+            result["response"] = response_content
+            result["messages"].append({"role": "assistant", "content": response_content})
+            
+            if message.get('tool_calls'):
+                result["tool_calls"] = message['tool_calls']
+            
+            # Handle JSON format if specified
+            if format == "json":
+                try:
+                    if isinstance(response_content, str):
+                        if response_content.startswith("```json"):
+                            response_content = (
+                                response_content.replace("```json", "")
+                                .replace("```", "")
+                                .strip()
+                            )
+                        parsed_response = json.loads(response_content)
+                        result["response"] = parsed_response
+                except json.JSONDecodeError:
+                    result["error"] = f"Invalid JSON response: {response_content}"
+            
+            return result
+
+    # Only if auto_process_tool_calls=True AND we have tools
+    # Make initial non-streaming call to check for tool calls
+    res = ollama.chat(**api_params, options=options)
     result["raw_response"] = res
     
-    # Extract the response content
-    response_content = res.get("message", {}).get("content")
-    result["response"] = response_content
+    #print("Raw Ollama response:", res)
+    
+    message = res.get("message", {})
+    response_content = message.get("content", "")
+    
+    # Check for tool calls and process them
+    if message.get('tool_calls'):
+        print("Found tool calls, processing automatically:", message['tool_calls'])
         
-    # Append response to messages
-    result["messages"].append({"role": "assistant", "content": response_content})
-
-
-    if tools and hasattr(res.get('message', {}), 'tool_calls') and res['message']['tool_calls']:
-        if tool_map:
-            response_for_processing = {
-                "response": res['message'].get('content'),
-                "raw_response": res,
-                "messages": messages,
-                "tool_calls": res['message']['tool_calls']
+        result["tool_calls"] = message['tool_calls']
+        
+        response_for_processing = {
+            "response": response_content,
+            "raw_response": res,
+            "messages": messages,
+            "tool_calls": message['tool_calls']
+        }
+        
+        # Process tool calls and get the updated result
+        processed_result = process_tool_calls(response_for_processing, tool_map, model, 'ollama', messages, stream=False)
+        
+        # Now if streaming was requested, make a final call with the complete conversation
+        if stream:
+            print("Making final streaming call with processed tools")
+            
+            # Use the updated messages from tool processing
+            final_messages = processed_result["messages"]
+            
+            # Make the final streaming call
+            final_api_params = {
+                "model": model,
+                "messages": final_messages,
+                "stream": True,
             }
-            return process_tool_calls(response_for_processing, tool_map, model, 'ollama', messages, stream)
-
-    # Handle JSON format if specified
-    if format == "json":
-        try:
-            if isinstance(response_content, str):
-                if response_content.startswith("```json"):
-                    response_content = (
-                        response_content.replace("```json", "")
-                        .replace("```", "")
-                        .strip()
-                    )
-                parsed_response = json.loads(response_content)
-                result["response"] = parsed_response
-        except json.JSONDecodeError:
-            result["error"] = f"Invalid JSON response: {response_content}"
-
-    return result
+            
+            if tools:
+                final_api_params["tools"] = tools
+            
+            final_stream = ollama.chat(**final_api_params, options=options)
+            processed_result["response"] = final_stream
+            
+        return processed_result
+    
+    # No tool calls found, handle normally
+    else:
+        result["response"] = response_content
+        result["messages"].append({"role": "assistant", "content": response_content})
+        
+        if stream:
+            # Make streaming call for regular response
+            stream_api_params = {
+                "model": model,
+                "messages": messages,
+                "stream": True,
+            }
+            if tools:
+                stream_api_params["tools"] = tools
+            
+            result["response"] = ollama.chat(**stream_api_params, options=options)
+        else:
+            # Handle JSON format if specified
+            if format == "json":
+                try:
+                    if isinstance(response_content, str):
+                        if response_content.startswith("```json"):
+                            response_content = (
+                                response_content.replace("```json", "")
+                                .replace("```", "")
+                                .strip()
+                            )
+                        parsed_response = json.loads(response_content)
+                        result["response"] = parsed_response
+                except json.JSONDecodeError:
+                    result["error"] = f"Invalid JSON response: {response_content}"
+        
+        return result
 
 
 def get_litellm_response(
@@ -236,6 +309,8 @@ def get_litellm_response(
     api_url: str = None,
     stream: bool = False,
     attachments: List[str] = None,
+    auto_process_tool_calls: bool = False,  # New kwarg
+
     **kwargs,
 ) -> Dict[str, Any]:
     result = {
@@ -246,11 +321,13 @@ def get_litellm_response(
         "tool_results":[],
     }
     if provider == "ollama":
-        kwargs["tool_map"] = tool_map
         return get_ollama_response(
-            prompt, model, images=images, tools=tools, tool_choice=tool_choice, think=think,
-            format=format, messages=messages, stream=stream, attachments=attachments, **kwargs
+            prompt, model, images=images, tools=tools, tool_choice=tool_choice, tool_map=tool_map, think=think,
+            format=format, messages=messages, stream=stream, attachments=attachments, 
+            auto_process_tool_calls=auto_process_tool_calls, **kwargs
         )
+    
+
     if attachments:
         for attachment in attachments:
             if os.path.exists(attachment):
@@ -339,6 +416,7 @@ def get_litellm_response(
 
 
     
+
     api_params = {"messages": result["messages"]}
     
     if api_url is not None and provider == "openai-like":
@@ -366,53 +444,115 @@ def get_litellm_response(
         for key, value in kwargs.items():
             if key in [
                 "stop", "temperature", "top_p", "max_tokens", "max_completion_tokens",
-                "tools", "tool_choice", "extra_headers", "parallel_tool_calls",
+                 "extra_headers", "parallel_tool_calls",
                 "response_format", "user",
             ]:
                 api_params[key] = value
-    
-    if tools and tool_map:
-        resp = completion(**{**api_params, "stream": False})
-        result["raw_response"] = resp
-        if hasattr(resp.choices[0].message, 'tool_calls') and resp.choices[0].message.tool_calls:
-            result["tool_calls"] = resp.choices[0].message.tool_calls
-            return process_tool_calls(result, tool_map, model, provider, messages, stream)
-    
-    api_params["stream"] = stream
-    #print(api_params)
-    resp = completion(**api_params)
 
-    if stream:
-        result["response"] = resp
-    else:
+    # If we want raw tool calls OR no tools, just call directly with requested streaming
+    if not auto_process_tool_calls or not (tools and tool_map):
+        api_params["stream"] = stream
+        resp = completion(**api_params)
         result["raw_response"] = resp
+        
+        if stream:
+            result["response"] = resp  # This is the stream generator
+            return result
+        else:
+            # Non-streaming response
+            llm_response = resp.choices[0].message.content
+            result["response"] = llm_response
+            result["messages"].append({"role": "assistant", "content": llm_response})
+            
+            # Check for tool calls
+            if hasattr(resp.choices[0].message, 'tool_calls') and resp.choices[0].message.tool_calls:
+                result["tool_calls"] = resp.choices[0].message.tool_calls
+            
+            # Handle JSON format requests
+            if format == "json":
+                try:
+                    if isinstance(llm_response, str):
+                        if llm_response.startswith("```json"):
+                            llm_response = llm_response.replace("```json", "").replace("```", "").strip()
+                        parsed_json = json.loads(llm_response)
+                        
+                        if "json" in parsed_json:
+                            result["response"] = parsed_json["json"]
+                        else:
+                            result["response"] = parsed_json
+                        
+                except (json.JSONDecodeError, TypeError) as e:
+                    print(f"JSON parsing error: {str(e)}")
+                    print(f"Raw response: {llm_response}")
+                    result["error"] = "Invalid JSON response"
+            
+            return result
+
+    # Only if process_tool_calls=True AND we have tools
+    # Make initial non-streaming call to check for tool calls
+    initial_api_params = api_params.copy()
+    initial_api_params["stream"] = False
+    
+    print("Initial API params:", initial_api_params)
+    resp = completion(**initial_api_params)
+    result["raw_response"] = resp
+    
+    # Check for tool calls
+    has_tool_calls = hasattr(resp.choices[0].message, 'tool_calls') and resp.choices[0].message.tool_calls
+    
+    if has_tool_calls:
+        print("Found tool calls in LiteLLM, processing automatically:", resp.choices[0].message.tool_calls)
+        
+        result["tool_calls"] = resp.choices[0].message.tool_calls
+        
+        # Process tool calls
+        processed_result = process_tool_calls(result, tool_map, model, provider, result["messages"], stream=False)
+        
+        # If streaming was requested, make final streaming call with processed conversation
+        if stream:
+            print("Making final streaming call with processed tools")
+            
+            final_api_params = api_params.copy()
+            final_api_params["messages"] = processed_result["messages"]
+            final_api_params["stream"] = True
+            
+            final_stream = completion(**final_api_params)
+            processed_result["response"] = final_stream
+            
+        return processed_result
+    
+    # No tool calls found, handle normally
+    else:
         llm_response = resp.choices[0].message.content
         result["response"] = llm_response
         result["messages"].append({"role": "assistant", "content": llm_response})
+        
+        if stream:
+            # Make streaming call for regular response
+            stream_api_params = api_params.copy()
+            stream_api_params["stream"] = True
+            final_stream = completion(**stream_api_params)
+            result["response"] = final_stream
+        else:
+            # Handle JSON format requests for non-streaming
+            if format == "json":
+                try:
+                    if isinstance(llm_response, str):
+                        if llm_response.startswith("```json"):
+                            llm_response = llm_response.replace("```json", "").replace("```", "").strip()
+                        parsed_json = json.loads(llm_response)
+                        
+                        if "json" in parsed_json:
+                            result["response"] = parsed_json["json"]
+                        else:
+                            result["response"] = parsed_json
+                        
+                except (json.JSONDecodeError, TypeError) as e:
+                    print(f"JSON parsing error: {str(e)}")
+                    print(f"Raw response: {llm_response}")
+                    result["error"] = "Invalid JSON response"
 
-    # Handle JSON format requests
-    if format == "json":
-        try:
-            if isinstance(llm_response, str):
-                # Clean up JSON response if needed
-                if llm_response.startswith("```json"):
-                    llm_response = llm_response.replace("```json", "").replace("```", "").strip()
-                parsed_json = json.loads(llm_response)
-                
-                if "json" in parsed_json:
-                    result["response"] = parsed_json["json"]
-                else:
-                    result["response"] = parsed_json
-                
-        except (json.JSONDecodeError, TypeError) as e:
-            print(f"JSON parsing error: {str(e)}")
-            print(f"Raw response: {llm_response}")
-            result["error"] = "Invalid JSON response"
-    
-    
-
-    return result
-
+        return result
 
 def process_tool_calls(response_dict, tool_map, model, provider, messages, stream=False):
     result = response_dict.copy()
@@ -425,11 +565,12 @@ def process_tool_calls(response_dict, tool_map, model, provider, messages, strea
     
     if not tool_calls:
         return result
-
+    print('tm', tool_map)
     for tool_call in tool_calls:
         tool_id = str(uuid.uuid4())
         tool_name = None
         arguments = {}
+        print('tc', tool_call)
 
         if isinstance(tool_call, dict):
             tool_id = tool_call.get("id", str(uuid.uuid4()))
@@ -448,14 +589,18 @@ def process_tool_calls(response_dict, tool_map, model, provider, messages, strea
             arguments = json.loads(arguments_str) if isinstance(arguments_str, str) else arguments_str
         except json.JSONDecodeError:
             arguments = {"raw_arguments": arguments_str}
-
+        print('arg', arguments)
+        print('tool name in tool map ', tool_name in tool_map)
         if tool_name in tool_map:
             tool_result = None
             tool_result_str = ""
             serializable_result = None
 
             try:
+                print(tool_map[tool_name])
+                print('Executing tool:', tool_name, 'with arguments:', arguments)
                 tool_result = tool_map[tool_name](**arguments)
+                print('Executed Tool Result:', tool_result)
             except Exception as e:
                 tool_result = f"Error executing tool '{tool_name}': {str(e)}"
 
@@ -485,7 +630,7 @@ def process_tool_calls(response_dict, tool_map, model, provider, messages, strea
                         "type": "function",
                         "function": {
                             "name": tool_name,
-                            "arguments": json.dumps(arguments)
+                            "arguments": arguments
                         }
                     }
                 ]
