@@ -19,6 +19,7 @@ from npcpy.gen.response import get_litellm_response
 from npcpy.gen.image_gen import generate_image, edit_image
 from npcpy.gen.video_gen import generate_video_diffusers
 
+
 def gen_image(
     prompt: str,
     model: str = None,
@@ -154,6 +155,7 @@ def get_llm_response(
             model = "llama3.2"
             
     system_message = get_system_message(npc) if npc else "You are a helpful assistant."
+    #print(system_message)
 
     if context is not None:
         context_str = f'User Provided Context: {context}'
@@ -326,6 +328,48 @@ def execute_llm_command(
         "messages": messages,
         "output": "Max attempts reached. Unable to execute the command successfully.",
     }
+def check_llm_command(
+    command: str,
+    model: str = None,
+    provider: str = None,
+    api_url: str = None,
+    api_key: str = None,
+    npc: Any = None,
+    team: Any = None,
+    messages: List[Dict[str, str]] = None,
+    images: list = None,
+    stream=False,
+    context=None,
+    shell=False,
+    actions: Dict[str, Dict] = None,
+    max_iterations: int = 3,
+    current_iteration: int = 0,
+):
+    """This function checks an LLM command and returns sequences of steps with parallel actions."""
+    if messages is None:
+        messages = []
+
+    if actions is None:
+        actions = DEFAULT_ACTION_SPACE.copy()
+
+    return execute_multi_step_plan(
+        command=command,
+        model=model,
+        provider=provider,
+        api_url=api_url,
+        api_key=api_key,
+        npc=npc,
+        team=team,
+        messages=messages,
+        images=images,
+        stream=stream,
+        context=context,
+        shell=shell,
+        actions=actions,
+        max_iterations=max_iterations,
+        current_iteration=current_iteration,
+        completed_actions=[]  # Initialize empty completed actions
+    )
 
 def handle_jinx_call(
     command: str,
@@ -336,6 +380,7 @@ def handle_jinx_call(
     api_key: str = None,
     messages: List[Dict[str, str]] = None,
     npc: Any = None,
+    team: Any = None,
     stream=False,
     shell=False,
     n_attempts=3,
@@ -357,15 +402,22 @@ def handle_jinx_call(
         the jinx call.
 
     """
-    if npc is None:
+    if npc is None and team is None:
         return f"No jinxs are available. "
     else:
-        if jinx_name not in npc.jinxs_dict:
+
+        #print(npc, team)
+        #print(team.jinxs_dict, npc.jinxs_dict)
+        if jinx_name not in npc.jinxs_dict and jinx_name not in team.jinxs_dict:
             print("not available")
-            print(f"jinx '{jinx_name}' not found in NPC's jinxs_dict.")
-            return f"jinx '{jinx_name}' not found."
+
+            return  {'output': f"jinx '{jinx_name}' not found in NPC's jinxs_dict.",
+                    'messages': messages}
         elif jinx_name in npc.jinxs_dict:
             jinx = npc.jinxs_dict[jinx_name]
+        elif jinx_name in team.jinxs_dict:
+            jinx = team.jinxs_dict[jinx_name]
+
         render_markdown(f"jinx found: {jinx.jinx_name}")
         jinja_env = Environment(loader=FileSystemLoader("."), undefined=Undefined)
         example_format = {}
@@ -460,6 +512,7 @@ def handle_jinx_call(
                     provider=provider,
                     messages=messages,
                     npc=npc,
+                    team=team,
                     api_url=api_url,
                     api_key=api_key,
                     stream=stream,
@@ -480,7 +533,6 @@ def handle_jinx_call(
                 input_values,
                 jinja_env,
                 npc=npc,
-
                 messages=messages,
             )
             if 'llm_response' in jinx_output and 'messages' in jinx_output:
@@ -498,6 +550,7 @@ def handle_jinx_call(
                     provider=provider,
                     messages=messages,
                     npc=npc,
+                    team=team,
                     api_url=api_url,
                     api_key=api_key,
                     stream=stream,
@@ -517,6 +570,7 @@ def handle_jinx_call(
                         provider=provider,
                         messages=messages,
                         npc=npc,
+                        team=team,
                         api_url=api_url,
                         api_key=api_key,
                         stream=stream,
@@ -530,8 +584,6 @@ def handle_jinx_call(
                         "messages": messages,
                     }
 
-        # process the jinx call
-        #print(messages)
         if not stream and messages[-1]['role'] != 'assistant':
             # if the jinx has already added a message to the output from a final prompt we dont want to double that
             
@@ -600,7 +652,6 @@ def handle_request_input(
     )
     return user_input
 def jinx_handler(command, extracted_data, **kwargs):
-    print(f"JINX HANDLER: calling handle_jinx_call with jinx_name='{extracted_data.get('jinx_name')}'")
     return handle_jinx_call(
         command, 
         extracted_data.get('jinx_name'),
@@ -610,6 +661,7 @@ def jinx_handler(command, extracted_data, **kwargs):
         api_key=kwargs.get('api_key'),
         messages=kwargs.get('messages'),
         npc=kwargs.get('npc'),
+        team = kwargs.get('team'),
         stream=kwargs.get('stream'),
         shell=kwargs.get('shell'),
         context=kwargs.get('context')
@@ -617,7 +669,7 @@ def jinx_handler(command, extracted_data, **kwargs):
 
 def answer_handler(command, extracted_data, **kwargs):
     print(f"ANSWER HANDLER: cmd='{command}', extracted_data={extracted_data}")
-    return get_llm_response(
+    response =  get_llm_response(
         f"""
         Here is the user question: {command}
         The action chosen by the agent is: answer_question
@@ -633,17 +685,19 @@ def answer_handler(command, extracted_data, **kwargs):
         api_key=kwargs.get('api_key'),
         messages=kwargs.get('messages'),
         npc=kwargs.get('npc'),
-        stream=kwargs.get('stream'),
+        stream=False,
         images=kwargs.get('images')
     )
-
+    
+    return response
+    
 # Define `DEFAULT_ACTION_SPACE`
 
 DEFAULT_ACTION_SPACE = {
     "invoke_jinx": {
         "description": "Invoke a jinx (jinja-template execution script)",
         "handler": jinx_handler,
-        "context": lambda npc=None, **_: (
+        "context": lambda npc=None, team=None, **_: (
             """
             Use jinxs when appropriate. For example:
 
@@ -657,6 +711,15 @@ DEFAULT_ACTION_SPACE = {
 
             You MUST use a jinx if the request directly refers to a tool the AI cannot handle directly (e.g., 'run', 'open', 'search', etc).
 
+            To invoke a jinx, return the action 'invoke_jinx' along with the jinx specific name. 
+            An example for a jinx-specific return would be
+            {
+                "action": "invoke_jinx",
+                "jinx_name": "file_reader",
+                "explanation": "Read the contents of <full_filename_path_from_user_request> and <detailed explanation of how to accomplish the problem outlined in the request> ."
+            }
+            Do not use the jinx names as the action keys.  You must use the action 'invoke_jinx' to invoke a jinx!
+
             Here are the currently available jinxs:
             """ +
             (
@@ -666,6 +729,13 @@ DEFAULT_ACTION_SPACE = {
                 )
                 if npc and getattr(npc, "jinxs_dict", None)
                 else "  [No jinxs available for this NPC]"
+            ) +  "\n \n \n Team Jinxs: " + (
+                "\n".join(
+                    f"- {name}: {jinx.description}"
+                    for name, jinx in getattr(team, "jinxs_dict", {}).items()
+                )
+                if team and getattr(team, "jinxs_dict", None)
+                else "  [No jinxs available for this team]"
             )
         ),
         "output_keys": {
@@ -682,96 +752,310 @@ DEFAULT_ACTION_SPACE = {
         "output_keys": {}
     }
 }
-
-# check_llm_command implementation
-def check_llm_command(
-    command: str,
-    model: str = None, 
-    provider: str = None, 
-    api_url: str = None,
-    api_key: str = None,
-    npc: Any = None,
-    team: Any = None,
-    messages: List[Dict[str, str]] = None,
-    images: list = None,
-    stream=False,
-    context=None,    
-    shell=False,
-    actions: Dict[str, Dict] = None,
+def plan_multi_step_actions(
+   command: str,
+   actions: Dict[str, Dict],
+   npc: Any = None,
+   model: str = None,
+   provider: str = None,
+   api_url: str = None,
+   api_key: str = None,
+   context: str = None,
+   messages: List[Dict[str, str]] = None,
+   team: Any = None
 ):
-    """This function checks an LLM command."""
-    if messages is None:
-        messages = []
+    """
+    Analyzes the user's command and creates a complete, sequential plan of actions
+    by dynamically building a prompt from the provided action space.
+    """
+    
+    # --- Start of the corrected logic ---
+    
+    # 1. Build the prompt dynamically.
+    prompt = f"""
+Analyze the user's request: "{command}"
 
-    # Use default action space if none provided
-    if actions is None:
-        actions = DEFAULT_ACTION_SPACE.copy()
+Your task is to create a complete, sequential JSON plan to fulfill the entire request.
+Use the following context about available actions and tools to construct the plan.
 
-        # Add request_input if in shell mode
-        if shell:
-            actions["request_input"] = {
-                "description": "Request additional input from the user",
-                "handler": lambda cmd, extracted_data, **kwargs: handle_request_input(
-                    f"Explanation from check_llm_command: {extracted_data.get('explanation', '')} \n for the user input command: {cmd}",
-                    kwargs.get('model'), 
-                    kwargs.get('provider')
-                ),
-                "context": "Use this when the user's request is ambiguous or needs clarification.",
-                "output_keys": {}
-            }
+"""
 
-    prompt = f"A user submitted this query: {command}\n"
-
-    # Add ALL action contexts to prompt
+    # 2. Dynamically add context from the action space, executing lambdas as needed.
     for action_name, action_info in actions.items():
         ctx = action_info.get("context")
+        # If the context is a function (our lambda), call it to get the dynamic string.
         if callable(ctx):
             try:
-                ctx = ctx(npc=npc, team=team, context=context)
+                # Pass the npc object to the lambda so it can access the jinxs_dict.
+                ctx = ctx(npc=npc, team=team)
             except Exception as e:
                 print(f"[WARN] Failed to render context for action '{action_name}': {e}")
                 ctx = None
+        
         if ctx:
-            prompt += f"\n{ctx}"
-    
+            prompt += f"\n--- Context for action '{action_name}' ---\n{ctx}\n"
+
+    print(prompt)
+    # 3. Add the final instructions.
     prompt += f"""
-    These were the most recent 5 messages in the conversation which should help you respond appropriately:
+--- Instructions ---
+Based on the user's request and the context provided above, create a plan.
+The plan must be a JSON object with a single key, "actions". Each action must include:
+- "action": The name of the action to take.
+- "explanation": A clear description of the goal for this specific step.
 
-    {messages[-5:] if len(messages) > 5 else messages}
-
-    Respond with a JSON object containing:
-    - "action": one of {list(actions.keys())}
-    - "explanation": a brief explanation of why you chose this action.
-    """
-
-    # Build dynamic JSON format from ALL output_keys across ALL actions
-    all_output_keys = {key_name: key_info for action_info in actions.values() for key_name, key_info in action_info.get("output_keys", {}).items()}
-    
-    json_format_parts = ['"action": "' + '" | "'.join(actions.keys()) + '"']
-    json_format_parts.append('"explanation": "<your_explanation>"')
-    
-    for key_name, key_info in all_output_keys.items():
-        description = key_info.get("description", f"<{key_name}>")
-        json_format_parts.append(f'"{key_name}": "{description}"')
-    
-    json_format_str = "{\n    " + ",\n    ".join(json_format_parts) + "\n}"
-    
-    prompt += f"""
-    Return only the JSON object. Do not include any additional text.
-
-    The format of the JSON object is:
-    {json_format_str}
-
-    In your explanation, do not needlessly reference the user's files or provided context. Simply provide the explanation for your choice in as few words as possible.
-
-    Remember, the action must be one of {list(actions.keys())}. Do not use a jinx's name for the action.
-    """
-
-    if context:
-        prompt += f"\nAdditional relevant context from user:\n{context}\n"
+Example Request: "Find out who the current CEO of Microsoft is, then find their biography on Wikipedia"
+ 
+An Example Plan might look like this depending on the available actions:
+""" + """
+{
+  "actions": [
+    {
+      "action": "<action_name_1>",
+      "<action_specific_key_1..>": "<action_specific_value_1>",
+      <...> : ...,      
+      "explanation": "Identify the current CEO of Microsoft."
+    },
+    {
+      "action": "<action_name_2>",
+        "<action_specific_key_1..>": "<action_specific_value_1>",
+        "explanation": "Find the <action-specific> information identified in the previous step."
+    }
+  ]
+}
 
+"""+f"""
+Now, create the complete plan for the user's query: "{command}"
+Respond ONLY with the JSON plan.
+"""
 
     action_response = get_llm_response(
+        prompt,
+        model=model, provider=provider, api_url=api_url, api_key=api_key,
+        npc=npc, format="json", messages=[]
+    )
+
+    response_content = action_response.get("response", {})
+    if isinstance(response_content, str):
+        try:
+            response_content = json.loads(response_content)
+        except json.JSONDecodeError:
+            print(f"Error: Could not parse plan from LLM. Response: {response_content}")
+            return []
+    
+    return response_content.get("actions", [])
+
+def execute_multi_step_plan(
+   command: str,
+   model: str = None,
+   provider: str = None,
+   api_url: str = None,
+   api_key: str = None,
+   npc: Any = None,
+   team: Any = None,
+   messages: List[Dict[str, str]] = None,
+   images: list = None,
+   stream=False,
+   context=None,
+   shell=False,
+   actions: Dict[str, Dict] = None,
+   max_iterations: int = 3, # No longer used for recursion
+   current_iteration: int = 0, # No longer used for recursion
+   completed_actions: List = None # No longer used for recursion
+):
+    """
+    Creates a comprehensive plan and executes it sequentially, passing context
+    between steps for adaptive behavior.
+    """
+    
+    # 1. Get the complete plan upfront with the corrected call.
+    planned_actions = plan_multi_step_actions(
+        command=command,
+        actions=actions,
+        npc=npc,
+        model=model,
+        provider=provider,
+        api_url=api_url,
+        api_key=api_key,
+        context=context,
+        messages=messages,
+        team=team
+    )
+    
+    if not planned_actions:
+        print("Could not generate a multi-step plan. Answering directly.")
+        result = answer_handler(command=command, extracted_data={"explanation": "Answering the user's query directly."}, model=model, provider=provider, api_url=api_url, api_key=api_key, messages=messages, npc=npc, images=images)
+        return {"messages": result.get('messages', messages), "output": result.get('response')}
+
+    # 2. Execute the plan step-by-step.
+    step_outputs = []
+    current_messages = messages.copy()
+    print(planned_actions)
+    for i, action_data in enumerate(planned_actions):
+        print(f"--- Executing Step {i + 1} of {len(planned_actions)} ---")
+        print(f"Goal: {action_data.get('explanation')}")
+
+        action_name = action_data["action"]
+        handler = actions[action_name]["handler"]
+        
+        step_context = f"Context from previous steps: {json.dumps(step_outputs)}" if step_outputs else ""
+        step_command = action_data.get('explanation')
+
+        result = handler(
+           command=step_command, 
+           extracted_data=action_data,
+           model=model,
+           provider=provider, 
+           api_url=api_url,
+           api_key=api_key, 
+           messages=current_messages, 
+           npc=npc,
+           team=team,
+           stream=stream, 
+           shell=shell, 
+           context=step_context, 
+           images=images
+        )
+
+        action_output = result.get('output') or result.get('response')
+        step_outputs.append(action_output)
+        current_messages = result.get('messages', current_messages)
+
+    # 3. Compile the final result, passing the original command for context.
+    final_output = compile_sequence_results(
+       original_command=command,
+       outputs=step_outputs,
+       model=model, provider=provider, api_url=api_url,
+       api_key=api_key, npc=npc
+    )
+    return {"messages": current_messages, "output": final_output}
+
+def compile_sequence_results(original_command: str, outputs: List[str], model: str = None, provider: str = None, api_url: str = None, api_key: str = None, npc: Any = None) -> str:
+    """
+    Synthesizes a list of outputs from sequential steps into a single,
+    coherent final response, framed as an answer to the original query.
+    """
+    if not outputs:
+        return "The process completed, but produced no output."
+    if len(outputs) == 1:
+        return outputs[0]
+    
+    synthesis_prompt = f"""
+A user asked the following question:
+"{original_command}"
+
+To answer this, the following information was gathered in sequential steps:
+{json.dumps(outputs, indent=2)}
+
+Based *directly on the user's original question* and the information gathered, please
+provide a single, final, and coherent response. Answer the user's question directly.
+Do not mention the steps taken.
+
+Final Synthesized Response:
+"""
+    
+    try:
+        response = get_llm_response(
+            synthesis_prompt,
+            model=model, provider=provider, api_url=api_url,
+            api_key=api_key, npc=npc, messages=[]
+        )
+        synthesized = response.get("response", "")
+        if synthesized and synthesized.strip():
+            return synthesized
+    except Exception as e:
+        print(f"Error during final synthesis: {e}")
+    
+    return "\n\n".join(outputs)
+def compile_sequence_results(original_command: str, outputs: List[str], model: str = None, provider: str = None, api_url: str = None, api_key: str = None, npc: Any = None) -> str:
+    """
+    Synthesizes a list of outputs from sequential steps into a single,
+    coherent final response, framed as an answer to the original query.
+    """
+    
+    if not outputs:
+        return "The process completed, but produced no output."
+    
+    if len(outputs) == 1:
+        return outputs[0]
+    
+    # The prompt now includes the user's query, giving the LLM crucial context.
+    synthesis_prompt = f"""
+A user asked the following question:
+"{original_command}"
+
+To answer this, the following information was gathered in sequential steps:
+{json.dumps(outputs, indent=2)}
+
+Based *directly on the user's original question* and the information gathered, please
+provide a single, final, and coherent response. Answer the user's question directly.
+Do not mention the steps taken.
+
+Final Synthesized Response:
+"""
+    
+    try:
+        response = get_llm_response(
+            synthesis_prompt,
+            model=model,
+            provider=provider,
+            api_url=api_url,
+            api_key=api_key,
+            npc=npc,
+            messages=[],
+        )
+        
+        synthesized = response.get("response", "")
+        if synthesized and synthesized.strip():
+            return synthesized
+    except Exception as e:
+        print(f"Error during final synthesis: {e}")
+    
+    # Fallback to a simple join if synthesis fails.
+    return "\n\n".join(outputs)
+
+
+def should_continue_with_more_actions(
+    original_command: str,
+    completed_actions: List[Dict[str, Any]],
+    current_messages: List[Dict[str, str]],
+    model: str = None,
+    provider: str = None,
+    api_url: str = None,
+    api_key: str = None,
+    npc: Any = None,
+) -> Dict:
+    """Decide if more action sequences are needed."""
+    
+    results_summary = ""
+    for idx, action_result in enumerate(completed_actions):
+        action_name = action_result.get("action", "Unknown Action")
+        output = action_result.get('output', 'No Output')
+        output_preview = output[:100] + "..." if isinstance(output, str) and len(output) > 100 else output
+        results_summary += f"{idx + 1}. {action_name}: {output_preview}\n"
+
+    prompt = f"""
+Original user request: "{original_command}"
+
+This request asks for multiple things. Analyze if ALL parts have been addressed.
+Look for keywords like "and then", "use that to", "after that" which indicate multiple tasks.
+
+Completed actions so far:
+{results_summary}
+
+For the request "{original_command}", identify:
+1. What parts have been completed
+2. What parts still need to be done
+
+JSON response:
+{{
+    "needs_more_actions": true/false,
+    "reasoning": "explain what's been done and what's still needed",
+    "next_focus": "if more actions needed, what specific task should be done next"
+}}
+"""
+
+    response = get_llm_response(
         prompt,
         model=model,
         provider=provider,
@@ -782,60 +1066,9 @@ def check_llm_command(
         messages=[],
         context=None,
     )
-
-
-    if "Error" in action_response:
-        print(f"LLM Error: {action_response['error']}")
-        return action_response["error"]
-
-    response_content = action_response.get("response", {})
-
-    if isinstance(response_content, str):
-        try:
-            response_content_parsed = json.loads(response_content)
-        except json.JSONDecodeError as e:
-            print(f"Invalid JSON received from LLM: {e}. Response was: {response_content}")
-            return f"Error: Invalid JSON from LLM: {response_content}"
-    else:
-        response_content_parsed = response_content
-
-    # Extract action and base fields
-    action = response_content_parsed.get("action", "").strip()
-
-    if action not in actions:
-        return {"messages": messages, "output": f"Error: Unknown action '{action}'"}
-
-    # Extract ONLY the fields defined in the chosen action's output_keys
-    extracted_data = {"action": action, "explanation": response_content_parsed.get("explanation", "").strip()}
-    chosen_action_output_keys = actions[action].get("output_keys", {})
-
-    for key_name in chosen_action_output_keys.keys():
-        if key_name in response_content_parsed:
-            extracted_data[key_name] = response_content_parsed[key_name]
-
-    # Build display string dynamically
-    display_parts = [f"Action chosen: {action}"]
-    display_parts.append(f"Explanation given: {extracted_data.get('explanation', '')}")
-    for key_name, value in extracted_data.items():
-        if key_name not in ["action", "explanation"]:
-            display_parts.append(f"{key_name}: {value}")
     
-    render_markdown("- " + "\n- ".join(display_parts) + "\n")
-
-    # Execute the action using the handler
-    handler = actions[action]["handler"]
-    
-    # Just pass everything, let the handler deal with it
-    try:
-        result = handler(command, extracted_data,
-                         model=model, provider=provider, api_url=api_url,
-                         api_key=api_key, messages=messages, npc=npc,
-                         stream=stream, shell=shell, context=context, images=images)
-        if isinstance(result, dict):
-            return {'messages': result.get('messages', messages), 'output': result.get('output', result.get('response', ''))}
-        else:
-            return {'messages': messages, 'output': result}
-            
-    except Exception as e:
-        print(f"Error executing action '{action}': {e}")
-        return {"messages": messages, "output": f"Error executing action '{action}': {str(e)}"}
\ No newline at end of file
+    response_dict = response.get("response", {})
+    if not isinstance(response_dict, dict):
+        return {"needs_more_actions": False, "reasoning": "Error", "next_focus": ""}
+        
+    return response_dict
\ No newline at end of file
