@@ -18,6 +18,7 @@ from jinja2 import Environment, FileSystemLoader, Template, Undefined
 from sqlalchemy import create_engine, text
 import npcpy as npy 
 from npcpy.llm_funcs import DEFAULT_ACTION_SPACE
+from npcpy.tools import auto_tools
 
 from npcpy.npc_sysenv import (
     ensure_dirs_exist, 
@@ -93,8 +94,11 @@ class Jinx:
                 messages=None):
         """Execute the jinx with given inputs"""
         if jinja_env is None:
+            # For standalone jinx execution, we don't need FileSystemLoader
+            # Just use a basic environment with no file system dependencies
+            from jinja2 import DictLoader
             jinja_env = Environment(
-                loader=FileSystemLoader([npc.npc_directory, npc.jinxs_directory]),
+                loader=DictLoader({}),  # Empty dict loader since we're not loading templates from files
                 undefined=SilentUndefined,
             )
         # Create context with input values and jinxs
@@ -417,6 +421,7 @@ class NPC:
         file: str = None,
         name: str = None,
         primary_directive: str = None,
+        plain_system_message: bool = False,
         jinxs: list = None,
         tools: list = None,
         model: str = None,
@@ -458,11 +463,23 @@ class NPC:
             #for these cases
             # if npcsh is initialized, use the ~/.npcsh/npc_team
             # otherwise imply
-            self.jinxs_directory = os.path.expanduser('~/.npcsh/npc_team/jinxs/')
-
+            if use_global_jinxs:
+                self.jinxs_directory = os.path.expanduser('~/.npcsh/npc_team/jinxs/')
+            else: 
+                self.jinxs_directory = None
             self.npc_directory = None # only makes sense when the input is also a file 
             # keep the jinxs tho to enable easieros.path.abspath('./npc_team/')
 
+        if tools is not None:
+            tools_schema, tool_map = auto_tools(tools)
+            self.tools = tools_schema  # Always store OpenAI-compatible schema here
+            self.tool_map = tool_map   # Store the callable map
+            self.tools_schema = tools_schema  # For backward compatibility if needed
+        else:
+            self.tools = []
+            self.tool_map = {}
+            self.tools_schema = []
+        self.plain_system_message = plain_system_message
         self.use_global_jinxs = use_global_jinxs
         
         self.memory_length = 20
@@ -556,9 +573,10 @@ class NPC:
         # Set NPC-specific jinxs directory path
         self.npc_jinxs_directory = os.path.join(os.path.dirname(file), "jinxs")
     def get_system_prompt(self, simple=False):
-        if simple:
+        if simple or self.plain_system_message:
             return self.primary_directive
         else:
+                
             return get_system_message(self)
     def _setup_db(self):
         """Set up database tables and determine type"""
@@ -596,6 +614,9 @@ class NPC:
         """Load and process NPC-specific jinxs"""
         npc_jinxs = []
         
+        if self.jinxs_directory is None:
+            self.jinxs_dict = {}
+            return None
         # Handle wildcard case - load all jinxs from the jinxs directory
         if jinxs == "*":
             #print(f'loading all jinxs for {self.name}')
@@ -649,10 +670,15 @@ class NPC:
                          tool_map= None,
                          tool_choice=None, 
                          messages: Optional[List[Dict[str, str]]] = None,
+                         auto_process_tool_calls: bool = True,
                          **kwargs):
         """Get a response from the LLM"""
         
-        # Call the LLM
+        if tools is None:
+            if self.tools is not None:
+                tools = self.tools
+                tool_map = self.tool_map
+
         response = npy.llm_funcs.get_llm_response(
             request, 
             model=self.model, 
@@ -661,7 +687,8 @@ class NPC:
             jinxs=jinxs,
             tools=tools, 
             tool_map=tool_map,
-            tool_choice=tool_choice,            
+            tool_choice=tool_choice,           
+            auto_process_tool_calls=auto_process_tool_calls,
             messages=self.memory if messages is None else messages,
             **kwargs
         )        
@@ -721,11 +748,11 @@ class NPC:
             model=self.model,
             provider=self.provider,
             npc=self,
-            team=team,  # Even though this might not be passed through
-            messages=messages,
+            team=team,
+            messages=self.memory if messages is None else messages,
             context=context,
             stream=stream,
-            actions=actions  # Pass the NPC-specific action space with team closure
+            actions=actions  
         )
     
     def handle_agent_pass(self, 
@@ -756,16 +783,14 @@ class NPC:
             + "PLEASE CHOOSE ONE OF THE OTHER OPTIONS WHEN RESPONDING."
         )
         
-        # Pass to the target NPC - make sure to pass team context
+
         result = target_npc.check_llm_command(
             updated_command,
             messages=messages,
             context=target_npc.shared_context,
-            team=team,  # Pass team context through
+            team=team, 
             stream=stream
         )
-        
-        # Add metadata to track the pass
         if isinstance(result, dict):
             result['npc_name'] = target_npc.name
             result['passed_from'] = self.name
@@ -797,8 +822,15 @@ class NPC:
     
     def __str__(self):
         """String representation of NPC"""
-        return f"NPC: {self.name}\nDirective: {self.primary_directive}\nModel: {self.model}\nProvider: {self.provider}\nAPI URL: {self.api_url}\njinxs: {', '.join([jinx.jinx_name for jinx in self.jinxs])}"
-
+        str_rep = f"NPC: {self.name}\nDirective: {self.primary_directive}\nModel: {self.model}\nProvider: {self.provider}\nAPI URL: {self.api_url}\n"
+        if self.jinxs:
+            str_rep += "Jinxs:\n"
+            for jinx in self.jinxs:
+                str_rep += f"  - {jinx.jinx_name}\n"
+        else:
+            str_rep += "No jinxs available.\n"
+        return str_rep 
+    
 class Team:
     def __init__(self, 
                  team_path=None, 
@@ -971,6 +1003,9 @@ class Team:
                             api_key=forenpc_api_key,
                             api_url=forenpc_api_url,                            
                                                 )
+            self.forenpc = forenpc
+            self.npcs[forenpc.name] = forenpc
+            return forenpc
         return None
     def get_npc(self, npc_ref):
         """Get NPC by name or reference with hierarchical lookup capability"""
@@ -1010,7 +1045,6 @@ class Team:
         # Initial request goes to forenpc
         result = forenpc.check_llm_command(request,
             context=getattr(self, 'context', {}),
-            #shared_context=self.shared_context,
             team = self, 
         )
         
@@ -1194,10 +1228,6 @@ class Team:
             
         return True
 
-# ---------------------------------------------------------------------------
-# Pipeline Runner
-# ---------------------------------------------------------------------------
-
 class Pipeline:
     def __init__(self, pipeline_data=None, pipeline_path=None, npc_team=None):
         """Initialize a pipeline from data or file path"""
